/* 
   SquidTL - Time Limit for Squid proxy
   Copyright (C) 2009-2010, Michele Pinassi <o-zone@zerozone.it>.
   
   http://www.zerozone.it/Software/Linux/SquidTL
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "HTEscape.h"
#include <getopt.h>
#include "common.h"

#define MAX_BUF 4096

struct SquidInfo {
  char protocol[MAX_BUF];
  char domain[MAX_BUF];
  int  dot;  /* true if domain is in dot notation */
  char url[MAX_BUF];
  char orig[MAX_BUF];
  char surl[MAX_BUF];
  char furl[MAX_BUF];
  char *strippedurl;
  int  port;
  char src[MAX_BUF];
  char srcDomain[MAX_BUF];
  char ident[MAX_BUF];
  char method[MAX_BUF];
  char realident[MAX_BUF];
  char context[MAX_BUF];
  unsigned int dynamic;
};

#ifdef XMLRPC
xmlrpc_env rpcEnv;
#endif

/* Converts a hex character to its integer value */
char from_hex(char ch) {
  return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
}

/* Converts an integer value to its hex character*/
char to_hex(char code) {
  static char hex[] = "0123456789abcdef";
  return hex[code & 15];
}

/* Returns a url-encoded version of str */
/* IMPORTANT: be sure to free() the returned string after use */
char *url_encode(char *str) {
  char *pstr = str, *buf = malloc(strlen(str) * 3 + 1), *pbuf = buf;
  while (*pstr) {
    if (isalnum(*pstr) || *pstr == '-' || *pstr == '_' || *pstr == '.' || *pstr == '~') 
      *pbuf++ = *pstr;
    else if (*pstr == ' ') 
      *pbuf++ = '+';
    else 
      *pbuf++ = '%', *pbuf++ = to_hex(*pstr >> 4), *pbuf++ = to_hex(*pstr & 15);
    pstr++;
  }
  *pbuf = '\0';
  return buf;
}


int parseLine(char *line, struct SquidInfo *s) {
  char *p, *d = NULL, *a = NULL, *e = NULL, *o, *field;
  int i = 0;
  char c;
  int report_once = 1;
  size_t strsz;
  int ndx = 0;
  
  field = strtok(line,"\t ");
  /*field holds each fetched url*/
  /* Let's first decode the url and then test it. Fixes bug2. */
  HTUnEscape(field);

  if(field == NULL)
    return 0;
  strcpy(s->orig,field);
  /* Now convert url to lowercase chars */ 
  for(p=field; *p != '\0'; p++) {
    *p = tolower(*p);
  }
  s->url[0] = s->protocol[0] = s->domain[0] = s->src[0] = s->ident[0] = 
    s->method[0] = s->srcDomain[0] = s->surl[0] =  '\0';
  s->dot = 0;
  s->port = 0;
  p = strstr(field,"://");
  if(p == NULL) { /* no protocol, defaults to http */
    strcpy(s->protocol,"unknown");
    p = field;
  } else {
    strncpy(s->protocol,field,p - field);
    *(s->protocol + ( p - field)) = '\0';
    p+=3; /* JMC -- 3 == strlen("://") */
    /* Now p only holds the pure URI */
    /* Fix for multiple slash vulnerability (bug1). */
    /* Check if there are still two or more slashes in sequence which must not happen */
    strsz = strlen(p);

    /* loop thru the string 'p' until the char '?' is hit or the "end" is hit */
    while('?' != p[ndx] && '\0' != p[ndx])
    {
        /* in case this is a '://' skip over it, but try to not read past EOS */
        if(3 <= strsz-ndx) {
          if(':' == p[ndx] && '/' == p[ndx+1] && '/' == p[ndx+2]) {
           ndx+=3; /* 3 == strlen("://"); */
          }
        }
        
       /* if this char and the next char are slashes,
 *           then shift the rest of the string left one char */
       if('/' == p[ndx] && '/' == p[ndx+1]) {
         size_t sz = strlen(p+ndx+1);
         strncpy(p+ndx,p+ndx+1, sz);
         p[ndx+sz] = '\0';
          if(1 == report_once) {
            // sgLogError("Warning: Possible bypass attempt. Found multiple slashes where only one is expected: %s", s->orig); 
            report_once--;
          }
      }
      else
      {
        /* increment the string indexer */
	assert(ndx < strlen(p));
        ndx++;
      }
    }
  }

  i=0;
  d = strchr(p,'/'); /* find domain end */
  /* Check for the single URIs (d) */
  e = d;
  a = strchr(p,'@'); /* find auth  */
  if(a != NULL && ( a < d || d == NULL)) 
    p = a + 1;
  a = strchr(p,':'); /* find port */;
  if(a != NULL && (a < d || d == NULL)){
    o = a + strspn(a+1,"0123456789") + 1;
    c = *o;
    *o = '\0';
    s->port = atoi(a+1);
    *o = c;
    e = a;
  }
  o=p;
  strcpy(s->furl,p);
  if (p[0] == 'w' || p[0] == 'f' ) {
    if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||
	(p[0] == 'w' && p[1] == 'e' && p[2] == 'b') ||
	(p[0] == 'f' && p[1] == 't' && p[2] == 'p')) {
      p+=3;
      while (p[0] >= '0' && p[0] <= '9')
	p++;
      if (p[0] != '.')
	p=o; /* not a hostname */
      else
	p++;
    }
  }
  if(e == NULL){
    strcpy(s->domain,o);
    strcpy(s->surl,p);
  }
  else {
    strncpy(s->domain,o,e - o);
    strcpy(s->surl,p);
    *(s->domain + (e - o)) = '\0';
    *(s->surl + (e - p)) = '\0';
  }
  //strcpy(s->surl,s->domain);
  if(strspn(s->domain,".0123456789") == strlen(s->domain))
    s->dot = 1;
  if(d != NULL)
    strcat(s->surl,d);
  s->strippedurl = s->surl;

  while((p = strtok(NULL," \t\n")) != NULL){
    switch(i){
    case 0: /* src */
      o = strchr(p,'/');
      if(o != NULL){
	strncpy(s->src,p,o-p);
	strcpy(s->srcDomain,o+1);
	s->src[o-p]='\0';
	if(*s->srcDomain == '-') {
	  s->srcDomain[0] = '\0';
	}
      } else {
	strcpy(s->src,p);
      }
      break;
    case 1: /* ident */
      if(strcmp(p,"-")){
	strcpy(s->ident,p);
	for(p=s->ident; *p != '\0'; p++) /* convert ident to lowercase chars */
	  *p = tolower(*p);
      } else
	s->ident[0] = '\0';
      break;
    case 2: /* method */
      strcpy(s->method,p);
      break;
    }
    i++;
  }
  if(s->domain[0] == '\0') {
    LOG_Write(DEBUG,"Domain is NULL: %s", s->orig);
    return 0;
  }
  if(s->method[0] == '\0') {
    LOG_Write(DEBUG,"Method is NULL: %s", s->orig);
    return 0;
  }
  return 1;
}

unsigned int onShutdown = 0,onLockAll=0,useXmlRpc=0;

#define OVECCOUNT 30

int doRegExp(char *pattern, char *data) {
    int match=0;
#ifdef PCRE
#warning Use of PCRE REGEXP 
    pcre *re;
    const char *error;
    int erroffset;
    int ovector[OVECCOUNT];
    int rc;

    re = pcre_compile(pattern, 0, &error,&erroffset, NULL);
    if (re == NULL) {
	/* Compilation failed: print the error message and exit */
	LOG_Write(ERROR, "PCRE RegExp error at offset %d: %s\n", erroffset, error); /* Print the error */
	return -1;
    }

    rc = pcre_exec(re, NULL, data, strlen(data), 0, 0, ovector, OVECCOUNT);
    /* Matching failed: handle error cases */
    if(rc >= 0) {
	match = 1;
    }
    pcre_free(re);     /* Release memory used for the compiled pattern */
#else
#warning Use of POSIX REGEXP (Deprecated !)
    regex_t *regex;
    int err_no=0; /* For regerror() */
    regmatch_t *result;
    
    regex = (regex_t *) malloc(sizeof(regex_t));
    memset(regex, 0, sizeof(regex_t));
    if((err_no=regcomp(regex, pattern, 0))!=0) /* Compile the regex */ {
	size_t length; 
	char *buffer;
	length = regerror (err_no, regex, NULL, 0);
	buffer = malloc(length);
	regerror(err_no, regex, buffer, length);
	LOG_Write(ERROR, "RegExp error: %s\n", buffer); /* Print the error */
	free(buffer);
	regfree(regex);
	return -1;
    }
    size_t no_sub = regex->re_nsub+1; /* How many matches are there in a line? */

    if((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
	LOG_Write(ERROR, "Cannot allocate required memory !"); /* Print the error */
	return -1;
    }
    
    if(regexec(regex, data, no_sub, result, 0) == 0) { /* Ok, match ! */
	match++;
    }

    regfree(regex); /* Free the regular expression data structure */
    free(regex);
#endif
    LOG_Write(DEBUG,"doRegExp: %s %s = %d\n",pattern,data,match);
    return match;
}

#define TRIM_CHARS  " \f\n\r\t\v"

char *trim(char *string) {
    char *result = NULL, *ptr = NULL;
    /* Ignore NULL pointers.  */
    if (string)
    {
        ptr = string;
        /* Skip leading whitespace.  */
        while (strchr(TRIM_CHARS, *ptr))
            ++ptr;
        /* Make a copy of the remainder.  */
        result = strdup(ptr);
        /* Move to the last character of the copy (that is, until the NULL char is reached).  */
        for (ptr = result; *ptr; ptr++);
        /* Move to the last non-whitespace character of the copy.  */
        for (--ptr; strchr(TRIM_CHARS, *ptr); --ptr);
        /* Remove trailing whitespace. */
        *(++ptr) = '\0';
    }
    return result;
}

int getRand() {
    return 1 + (int) (10.0 * (rand() / (RAND_MAX + 1.0)));
}

MYSQL *conn;
MYSQL_RES *res,*resb,*resc;
MYSQL_ROW row,rowb,rowc;

char buf[MAX_BUF],tmp[MAX_BUF];
struct _Config *Config;
char xmlrpcRepo[255];

void CB_Signal(int sig) {
    LOG_Write(DEBUG,"CB_Signal: %d",sig);
    switch(sig) {
	case SIGALRM:
	    /* Purge expiored accesses */
	    sprintf(tmp,"DELETE FROM DB_Access WHERE TIMESTAMPDIFF(MINUTE,NOW(),AddDate) > %d;",Config->purgeTimeout);
	    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
	    if(mysql_query(conn, tmp)) {
	        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
	    }
	    if(mysql_affected_rows(conn) > 0) {
		LOG_Write(NOTICE,"Purged %d old entries (Timeout: %d minutes)\n",mysql_affected_rows(conn),Config->purgeTimeout);
	    }
#ifdef XMLRPC
	    if(getXmlrpcRepo(&rpcEnv) > 0) {
		LOG_Write(NOTICE,"XML-RPC getRepository(MAIN): %s",xmlrpcRepo);
		useXmlRpc = 1;
	    } else {
		LOG_Write(WARNING,"XML-RPC getRepository(MAIN) FAILED !");
		useXmlRpc = 0;
	    }
#endif
	    break;
	case SIGHUP:
	case SIGSTOP:
	case SIGKILL:
	    if(!onShutdown) {
		onShutdown = 1;
		LOG_Write(NOTICE,"Shutdown request");
	    }
	    break;
	default:
	    break;
    }
    alarm(355+getRand());  /* Trigger ogni 360 secondi (10 random) = 5 minuti */
}

unsigned int checkTimeRule(int ruleId) {
    time_t t;
    struct tm * timeinfo;
    unsigned int retval=0;
    /* Return 1 if VALID, 0 if NOT */
    
    time(&t);
    timeinfo = localtime(&t);
    
    /*
    tm_sec	seconds after the minute	0-61*
    tm_min	minutes after the hour	0-59
    tm_hour	hours since midnight	0-23
    tm_mday	day of the month	1-31
    tm_mon	months since January	0-11
    tm_year	years since 1900	
    tm_wday	days since Sunday	0-6
    tm_yday	days since January 1	0-365
    tm_isdst	Daylight Saving Time flag
    */
    
    sprintf(tmp,"SELECT t2.ID,t2.Name,t2.fromHour,t2.toHour,t2.isNegate,t2.daysOfWeek FROM RulesTime AS t1 INNER JOIN DB_Time AS t2 ON t1.TimeID=t2.ID WHERE t1.RuleID=%d AND IF(t2.isNegate,(CURTIME() < t2.fromHour AND CURTIME() > t2.toHour),(t2.fromHour < CURTIME() AND t2.toHour > CURTIME()));",ruleId);
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
	LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
    } else {
	resc = mysql_store_result(conn);
	while((rowc = mysql_fetch_row(resc)) != NULL) {
	 /* Now check the DAYs... */
	    if(rowc[5][timeinfo->tm_wday] == '1') {
		LOG_Write(DEBUG,"Apply %s\n",rowc[1]);
		retval = 1;
	    }
	}
	mysql_free_result(resc);
    }
    LOG_Write(DEBUG,"checkTimeRule(%d) = %d\n",ruleId,retval);
    return retval;
}

/**************************************
checkDomainCategory

Query XML-RPC service to get URL category, if exists. Return 0 if don't match, in case of error or xml-rpc NOT active. Return 1 if match.
***************************************/
int checkDomainCategory(xmlrpc_env *env,char *category,char *domain) {
    char *resultString;
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    xmlNode *cur_node = NULL,*cur_node_a = NULL,*cur_node_b = NULL;
    int match = 0;

    LIBXML_TEST_VERSION
#ifdef XMLRPC
    if(useXmlRpc == 1) {
	resultString = (char *)XMLRPC_Query(env,xmlrpcRepo,"squidTl.getCategory",domain);
	if(resultString != NULL) {
	    LOG_Write(DEBUG,"XML-RPC getCategory(%s): %s",domain,resultString);
	    doc = xmlParseMemory(resultString, strlen(resultString));

	    if (doc == NULL) {
    		LOG_Write(ERROR,"getDomainCategory: Could not parse XMLRPC response !");
		return 0;
	    }

	    /*Get the root element node */
	    root_element = xmlDocGetRootElement(doc);

	    for(cur_node = root_element; cur_node; cur_node = cur_node->next) {
    		if(cur_node->type == XML_ELEMENT_NODE) {	    
    	    	    if(strcasecmp(cur_node->name,"message") == 0) {
    	    	    /* Got first level */
    	    		for(cur_node_a = cur_node->children; cur_node_a; cur_node_a = cur_node_a->next) {
    			    if(cur_node_a->type == XML_ELEMENT_NODE) {	    
    	    			if(strcasecmp(cur_node_a->name,"code") == 0) {
    	    			
    	    			}
    	    			if(strcasecmp(cur_node_a->name,"domain") == 0) {
    	    			
    	    			}
    	    			if(strcasecmp(cur_node_a->name,"category") == 0) {
    	    			    /* Got the URL for the repo ! */
    	    			    if(strncasecmp(category,xmlNodeGetContent(cur_node_a),255) == 0) {
    	    				match = 1;
					LOG_Write(DEBUG,"checkDomainCategory(%s,%s) = TRUE\n",domain,category);
    	    			    }
    	    			}
    	    		    }
    			}
		    }
		}
	    }
	    free(resultString);
	    return match;
	} else {
	    LOG_Write(ERROR,"Error while querying XML-RPC service @ %s\n",xmlrpcRepo);
	    return 0;
	}
    }
#else
    return 0;
#endif
}

/**************************************
getXmlrpcRepo

Query remote XML-RPC Hub for the URL of a REPO
***************************************/

int getXmlrpcRepo(xmlrpc_env *env) {
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    xmlNode *cur_node = NULL,*cur_node_a = NULL,*cur_node_b = NULL;
    char *resultString;

    LIBXML_TEST_VERSION

#ifdef XMLRPC
    resultString = (char *)XMLRPC_Query(env,Config->xmlrpcServer,"squidTl.getRepository","MAIN");
    if(resultString != NULL) {
        LOG_Write(DEBUG,"XML-RPC getXmlrpcRepo: %s",resultString);

	doc = xmlParseMemory(resultString,strlen(resultString));

	if (doc == NULL) {
    	    LOG_Write(ERROR,"getXmlrpcRepo: Could not parse XMLRPC response !");
	    return 0;
	}

	/*Get the root element node */
	root_element = xmlDocGetRootElement(doc);

	for(cur_node = root_element; cur_node; cur_node = cur_node->next) {
    	    if(cur_node->type == XML_ELEMENT_NODE) {	    
    	        if(strcasecmp(cur_node->name,"message") == 0) {
    	    	    /* Got first level */
    	    	    for(cur_node_a = cur_node->children; cur_node_a; cur_node_a = cur_node_a->next) {
    			if(cur_node_a->type == XML_ELEMENT_NODE) {	    
    	    		    if(strcasecmp(cur_node_a->name,"repo") == 0) {
    	    			/* Got second level */
    	    			for(cur_node_b = cur_node_a->children; cur_node_b; cur_node_b = cur_node_b->next) {
    				    if(cur_node_b->type == XML_ELEMENT_NODE) {
    	    				if(strcasecmp(cur_node_b->name,"url") == 0) {
    	    				    /* Got the URL for the repo ! */
    	    				    strncpy(xmlrpcRepo,xmlNodeGetContent(cur_node_b),255);
    	    				}
    	    			    }
    	    			}
    	    		    }
    			}
		    }
    		}
    	    }
	}
	xmlFreeDoc(doc);
	xmlCleanupParser();
	free(resultString);
	return 1;
    } else {
        LOG_Write(ERROR,"getXmlRpcRepo(): Error while querying XML-RPC service @ %s\n",Config->xmlrpcServer);
	return 0;
    }
#else
    return 0;
#endif
}

/**************************************
checkURLList

Check if the domain exists in specified URL list (text file). Return 0 if NOT found, 1 if found.
***************************************/
int checkURLList(char *urllist,char *domain) {
    FILE *fp;
    int match=0;
    sprintf(tmp,"%s/%s",Config->URLListPath,urllist);
    fp = fopen(tmp, "r");
    if(fp == NULL) {
        LOG_Write(ERROR,"Error %s while opening %s\n",strerror(errno),tmp);
        return -1;
    }
    while (!feof(fp)) {
	if(fgets(tmp, MAX_BUF, fp)) {
	    /* Match domain with url regexp */
	    if(doRegExp(trim(tmp),domain) > 0) {
		/* Yes, we got it ! Don't lose more time looking for other matches... */
		match = 1;
		break;
	    }
	}
    }
    fclose(fp);
    LOG_Write(NOTICE,"checkURLList(%s,%s) = %d\n",urllist,domain,match);
    return match;
}

int addNewUser(struct SquidInfo *s) {
    // Add user to DB_Users
    sprintf(tmp,"INSERT IGNORE INTO DB_Users (Username,Type,IdentifiedBy,Value,Note,AddDate) VALUES (\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",NOW());",s->ident,Config->defaultUserLevel,"IDENT",s->ident,"Added automatically by system based on unknown user action");
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
	LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
	return -1;
    }
    // Bind the new user to DB_Groups
    sprintf(tmp,"INSERT IGNORE INTO UserGroup (UserID,GroupID) VALUES (\"%s\",\"%d\");",s->ident,Config->defaultUserGroup);
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
	LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
	return -1;
    }
    return Config->defaultUserLevel;
}

/**************************************
getUserContext

get the user context based on a query to DB_Context and fill relative field in squidInfo
***************************************/

int getUserContext(struct SquidInfo *s) {
    sprintf(tmp,"SELECT ipRange,ID FROM DB_Context;");
    /* 
    ipRange = IP Range for this context
    ID = Context ID
    */
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
        return -1;
    }
    res = mysql_use_result(conn);
    while((row = mysql_fetch_row(res)) != NULL) {
        if(doRegExp(row[0],s->src) > 0) {
	    /* Ok, matched ! */
	    strncpy(s->context,row[1],MAX_BUF);
	    break;
	}
    }
    mysql_free_result(res);
    if(strlen(s->context) > 0) {
	LOG_Write(NOTICE,"getUserContext(%s) = %s\n",s->src,s->context);
	return 1;
    } else {
	LOG_Write(NOTICE,"getUserContext(%s) = DEFAULT CONTEXT\n",s->src);
	return 0;
    }
}

int getRealIdent(struct SquidInfo *s) {
    int level=-1;
    /* Return identity of source based on DB_Users */
    sprintf(tmp,"SELECT IdentifiedBy,Value,Username,Type FROM DB_Users");
    /* 
    Username = Identity of source
    IdentifiedBy = How i have to identify this source (IDENT|IP or range IP)
    Value = A value for this rule
    */
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
        return -1;
    }
    res = mysql_use_result(conn);
    while((row = mysql_fetch_row(res)) != NULL) {
        if((strcasecmp(row[0],"IP") == 0) && (doRegExp(row[1],s->src) > 0)) {
	    /* Ok, matched ! Because this is an IP matching, create unique user for each IP... */
	    sprintf(tmp,"%s_%s",row[2],s->src);
	    strncpy(s->realident,row[2],MAX_BUF); /* Realident riferisce al nome utente reale sul DB */
	    strncpy(s->ident,tmp,MAX_BUF); /* ident al nome univoco */
	    level = atoi(row[3]);
	    s->dynamic = 1;
	    break;
	}
        if((strcasecmp(row[0],"IDENT") == 0) && (doRegExp(row[1],s->ident) > 0)) {
	    /* Ok, matched ! */
	    strncpy(s->ident,row[2],MAX_BUF);
	    strncpy(s->realident,row[2],MAX_BUF);
	    level = atoi(row[3]);
	    s->dynamic = 0;
	    break;
	}
    }
    mysql_free_result(res);
    LOG_Write(NOTICE,"getRealIdent(%s) = %s (%d)\n",s->src,s->ident,level);
    return level;
}

int DB_Open() {
    conn = mysql_init(NULL);

    /* Connect to database */
    if (!mysql_real_connect(conn, Config->mysqlHost, Config->mysqlUser, Config->mysqlPassword, Config->mysqlDb, 0, NULL, 0)) {
	LOG_Write(FATAL,"MySQL Error: %s\n", mysql_error(conn));
	return -1;
    }

#if MYSQL_VERSION_ID > 50012
    mysql_options(conn, MYSQL_OPT_RECONNECT, "1");
#else
#warning NO AUTO-RECONNECT ENABLE !
#endif    
    return 0;
}

void DB_Close() {
    mysql_close(conn);
}

int main(int argc, char **argv) {
    struct SquidInfo squidInfo;
    /* ALARM Signal */
    sigset_t block, pause_mask;
    struct sigaction s;

    /* Default log file */
    char logfile[255];
    strcpy(logfile,"/tmp/squidtl.log\0");

    /* Default config file */
    char configfile[255];
    strcpy(configfile,"/etc/squidtl/config.xml\0");
    
    char tmpEnc[4096],redirectUrl[MAX_BUF];
    
    int rate=0,logVerbose=NOTICE,rulesId;
    
    int c,isAllowed;

    struct option long_options[] = {
               {"verbose",     no_argument,       &logVerbose, DEBUG},
               {"help",        no_argument,       0, 'h'},
               {"config",      required_argument, 0, 'c'},
               {"log",         required_argument, 0, 'l'},
               {0, 0, 0, 0}
    };

#ifdef XMLRPC
    XMLRPC_Init(&rpcEnv);
#endif

    while (1) {
	/* getopt_long stores the option index here. */
	int option_index = 0;
	c = getopt_long (argc, argv, "hc:l:", long_options, &option_index);
	/* Detect the end of the options. */
	if(c == -1) break;
	switch (c) {
	    case 'h':
		printf("Usage: %s [options]\n\n--h -help this help\n--verbose\n--c -config config file path\n-l --log log file path");
		exit(0);
	    case 'c':
		strncpy(configfile,optarg,255);
		break;
	    case 'l':
		strncpy(logfile,optarg,255);
		break;
	    default:
		break;
        }
    }
    
    LOG_Init(logfile,logVerbose);

    LOG_Write(NOTICE,"SquidTL v%s",VERSION);

    if((Config = (struct _Config *)CONFIG_Load(configfile)) == NULL) {
	LOG_Write(FATAL,"Config file error: exiting.");
	return(0);
    }

    alarm(360);  /* Trigger ogni 360 secondi = 5 minuti */
    signal(SIGALRM, CB_Signal);
    signal(SIGHUP, CB_Signal);
    signal(SIGSTOP, CB_Signal);
    signal(SIGKILL, CB_Signal);
    
    if(DB_Open() < 0) {
	LOG_Write(FATAL,"Cannot connect to MySQL DB !");
	return(0);
    }
    
    LOG_Write(NOTICE,"MySQL(%s:%s) connection OK",Config->mysqlHost,Config->mysqlUser);

#ifdef XMLRPC
    if(getXmlrpcRepo(&rpcEnv) > 0) {
	LOG_Write(NOTICE,"XML-RPC getRepository(MAIN): %s",xmlrpcRepo);
	useXmlRpc = 1;
    } else {
	LOG_Write(WARNING,"XML-RPC getRepository(MAIN) FAILED !");
    }
#endif
    while(!onShutdown) {
	while(fgets(buf, MAX_BUF, stdin) != NULL) {
	    LOG_Write(DEBUG,"IN: %s",buf);
	    if(parseLine(buf,&squidInfo) != 1) {
		/* On error always ALLOW */
		LOG_Write(ERROR,"Error parsing squid line: %s",buf);
		puts("");
	    } else if(onLockAll == 1) {
		/* DENY ALL if ALL TRAFFIC LOCKED (don't log those redirects) */
		LOG_Write(NOTICE,"ALL TRAFFIC LOCKED: BLOCK\n");
		sprintf(redirectUrl,"%s?rule=alllock",Config->defaultRedirectURL);
		fprintf(stdout,"302:%s\n",redirectUrl);
	    } else {
		/* Ok...check ! */
		int userLevel;
		isAllowed = 0;
		redirectUrl[0] = '\0';
		LOG_Write(DEBUG,"SquidInfo - P:%s D:%s U:%s O:%s S:%s F:%s\n",squidInfo.protocol, squidInfo.domain, squidInfo.url, squidInfo.orig, squidInfo.surl, squidInfo.furl);
		getUserContext(&squidInfo);
		userLevel = getRealIdent(&squidInfo); /* 0=Disabled, -1=Unknown >0=Identified */
		if(userLevel > 0) {
		    /* Add row to DB_Access */
	    	    mysql_real_escape_string(conn, tmpEnc, squidInfo.orig, strlen(squidInfo.orig));
		    sprintf(tmp,"INSERT INTO DB_Access (Src,Domain,FullUrl,Ident,AddDate) VALUES (\"%s\",\"%s\",\"%s\",\"%s\",NOW());",squidInfo.src,squidInfo.domain,tmpEnc,squidInfo.ident);
		    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		    if(mysql_query(conn, tmp)) {
			LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		    }
		    /* Now get groups owned by user (identified by realident) */
	    	    mysql_real_escape_string(conn, tmpEnc, squidInfo.realident, strlen(squidInfo.realident));
		    sprintf(tmp,"SELECT GroupID FROM UserGroup WHERE UserID='%s';",tmpEnc);
		    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		    if(mysql_query(conn, tmp)) {
			LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		    }
		    res = mysql_store_result(conn);
		    while((row = mysql_fetch_row(res)) != NULL) {
			/* Ok, now check all rules in those groups */
			sprintf(tmp,"SELECT t1.ID,t1.Type,t1.Value,t1.Action,t1.MaxHits,t1.TimeSlice,t1.redirectUrl FROM DB_Rules AS t1 INNER JOIN RulesGroup AS t2 ON t2.RuleID=t1.ID WHERE t2.GroupID='%s' AND t1.Context='%s';",row[0],squidInfo.context);
			/* 
			Type = Which type of rule is this ? (CATEGORY|URLLIST|DOMAIN|IDENT|IP)
			Value = A value for this rule
			Action = What i have to do when triggered ? (BLOCK)
			MaxHits = How many times...
			TimeSlice = ...in how many minutes ?
			redirectUrl = Where i have to redirect the bad guys ?
			*/
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(conn, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
			}
			resb = mysql_store_result(conn);
			while((rowb = mysql_fetch_row(resb)) != NULL) {
			    /* Check rule validity agains Time rules */
			    if(checkTimeRule(atoi(rowb[0])) > 0) { /* <!-- TIME VALIDITY --> */
				/* === CATEGORY THAT RELY ON XMLRPC === */
				if(strcasecmp(rowb[1],"CATEGORY") == 0) { /* <!-- CATEGORY --> */
#ifdef XMLRPC
				    if(checkDomainCategory(&rpcEnv,rowb[2],squidInfo.domain) > 0) {
					char *redirUrl, *urlencode;
					if(strncasecmp(rowb[3],"ALLOW",32) == 0) {
					    /* Explicit allow */
					    isAllowed = 1;
					} else {
					    /* Default: BLOCK */
					    if(strlen(rowb[6]) < 1) {
					        redirUrl = Config->defaultRedirectURL;
					    } else {
					        redirUrl = rowb[6];
					    }
					    urlencode = url_encode(rowb[2]);
					    sprintf(redirectUrl,"%s?rule=category&action=%s&category=%s&ident=%s",redirUrl,rowb[3],urlencode,squidInfo.ident);
					    free(urlencode);
					}
					rulesId = atoi(rowb[0]);
				    }
#else
				    LOG_Write(WARNING,"Category rules (ID:%d) work only with XMLRPC enabled !",atoi(rowb[0]));
#endif
				}
				/* === URLLIST BASED ON FILE === */
				if(strcasecmp(rowb[1],"URLLIST") == 0) { /* <!-- URLLIST --> */
				    if(checkURLList(rowb[2],squidInfo.domain) > 0) {
					char *redirUrl, *urlencode;
					if(strncasecmp(rowb[3],"ALLOW",32) == 0) {
					    /* Explicit allow */
					    isAllowed = 1;
					} else {
					    /* Default: BLOCK */
					    if(strlen(rowb[6]) < 1) {
					        redirUrl = Config->defaultRedirectURL;
					    } else {
					        redirUrl = rowb[6];
					    }
					    urlencode = url_encode(rowb[2]);
					    sprintf(redirectUrl,"%s?rule=urllist&action=%s&urllist=%s&ident=%s",redirUrl,rowb[3],urlencode,squidInfo.ident);
					    free(urlencode);
					}
					rulesId = atoi(rowb[0]);
				    }
				}
				/* === DOMAIN LIMITER === */
				if((strcasecmp(rowb[1],"DOMAIN") == 0) && (doRegExp(rowb[2],squidInfo.domain) > 0)) { /* <!-- DOMAIN --> */
				    /* Ok, now check the value */
				    sprintf(tmp,"SELECT Domain FROM DB_Access WHERE Ident=\"%s\" AND TIMESTAMPDIFF(MINUTE,AddDate,NOW()) < %s;",squidInfo.ident,rowb[5]);
				    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
				    if(mysql_query(conn, tmp)) {
					LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
				    } else {
					resc = mysql_store_result(conn);
					rate = 0;
					while((rowc = mysql_fetch_row(resc)) != NULL) {
					    if(doRegExp(rowb[2],rowc[0]) > 0) {
						rate++;
					    }
					}
					LOG_Write(DEBUG,"LIMIT for domain %s: %s hits in %s minutes. Ident %s have %d\n",rowb[2],rowb[4],rowb[5],squidInfo.ident,rate);
					if(rate > atoi(rowb[4])) {
					    char *redirUrl, *urlencode;
					    if(strncasecmp(rowb[3],"ALLOW",32) == 0) {
						/* Allow */
						isAllowed = 1;
					    } else {
						/* Default: BLOCK */
						if(strlen(rowb[6]) < 1) {
						    redirUrl = Config->defaultRedirectURL;
						} else {
						    redirUrl = rowb[6];
						}
						urlencode = url_encode(rowb[2]);
						sprintf(redirectUrl,"%s?rule=domain&action=%s&domain=%s&ident=%s&hits=%d&hitlimit=%s&timelimit=%s",redirUrl,rowb[3],urlencode,squidInfo.ident,rate,rowb[4],rowb[5]);
						free(urlencode);
					    }
					    rulesId = atoi(rowb[0]);
					}
					mysql_free_result(resc);
				    }
				} /* <!-- /DOMAIN --> */
			    } else { /* <!-- /TIME VALIDITY --> */
				LOG_Write(DEBUG,"Rule %s (%s=%s) ignored due to time rule\n",rowb[0],rowb[1],rowb[2]);
			    }
			}
			mysql_free_result(resb);
		    }
		    mysql_free_result(res);
		} else if(userLevel == 0) { /* Disabled user: always block */
		    LOG_Write(NOTICE,"DISABLED USER: BLOCK\n");
		    sprintf(redirectUrl,"%s?rule=disableduser",Config->defaultRedirectURL);
		} else {
		    /* Unknow user action */
		    LOG_Write(DEBUG,"UNKNOWN USER. Default action: %s\n",Config->defaultUserAction);
		    if(strncasecmp(Config->defaultUserAction,"BLOCK",32) == 0) {
			/* Block ALL unknown users */
			sprintf(redirectUrl,"%s?rule=unknownuser",Config->defaultRedirectURL);
		    } else if(strncasecmp(Config->defaultUserAction,"ADD",32) == 0) {
			if(strlen(squidInfo.ident) > 0) {
			    if(addNewUser(&squidInfo) > 0) {
				LOG_Write(NOTICE,"USER %s (%d) ADDED TO DB",squidInfo.ident,Config->defaultUserGroup);
			    } else {
				LOG_Write(ERROR,"Failed adding new user !");
			    }
			} else {
			    LOG_Write(ERROR,"Cannot identify unknown user. Automatic ADD is impossibile !");
			}
		    } else {
			/* In case of unknown action, pass ! */
			redirectUrl[0] = '\0';
		    }
		}
		/* Now check if needs to be redirected */
		if(strlen(redirectUrl) > 0) {
		    if(strncasecmp(squidInfo.orig,redirectUrl,MAX_BUF) == 0) {
			/* Prevent infinite loops */
			puts("");
		    } else {
			/* Oh oh, bad guy...needs to be redirected ! */
			fprintf(stdout,"302:%s\n",redirectUrl);
			LOG_Write(NOTICE,"REDIRECT: 302:%s\n",redirectUrl);
			/* Save event to DB for statistic purposes */
			sprintf(tmp,"INSERT INTO DB_Events (RulesID,Src,Ident,AddDate) VALUES (%d,\"%s\",\"%s\",NOW());",rulesId,squidInfo.src,squidInfo.ident);
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(conn, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
			}
		    }
		} else {
		    /* If not, check the default rule */
		    if(isAllowed == 1) {
			LOG_Write(NOTICE,"ALLOWED: %s (EXPLICIT)\n",squidInfo.domain);
			puts("");
		    } else {
			if(strncasecmp(Config->defaultAction,"BLOCK",32) == 0) {
			    sprintf(redirectUrl,"%s?rule=default&action=block",Config->defaultRedirectURL);
			    fprintf(stdout,"302:%s\n",redirectUrl);
			    LOG_Write(NOTICE,"REDIRECT: 302:%s (DEFAULT)\n",redirectUrl);
			} else {
			    LOG_Write(NOTICE,"ALLOWED: %s (DEFAULT)\n",squidInfo.domain);
			    puts("");
			}
		    }
		}
		break;
	    }
	}
	fflush(stdout);
	if(feof(stdin)) {
	    /* FEOF means that squid has gone away.... */
	    LOG_Write(FATAL,"FEOF on STDIN detected: Squid has gone away ?");
	    onShutdown = 1;
	}
    }
#ifdef XMLRPC
    XMLRPC_Close(&rpcEnv);
#endif
    LOG_Write(FATAL,"Shutdown NOW !");
    LOG_Close();
    DB_Close();
    exit(0);
}
