/* 
   SquidTL - Time Limit for Squid proxy
   Copyright (C) 2009-2010, Michele Pinassi <o-zone@zerozone.it>.
   
   http://www.zerozone.it/Software/Linux/SquidTL
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "common.h"

#ifdef XMLRPC

int XMLRPC_Init(xmlrpc_env *env) {
    xmlrpc_env_init(env);

    /* Start up our XML-RPC client library. */
    xmlrpc_client_init2(env, XMLRPC_CLIENT_NO_FLAGS, NAME, VERSION, NULL, 0);
    if(env->fault_occurred) {
        LOG_Write(ERROR, "XMLRPC_Init: %s (%d)\n", env->fault_string, env->fault_code);
        return 0;
    }
    return 1;
}

const char *XMLRPC_Query(xmlrpc_env *env,char *serverUrl,char *methodName,char *methodArgs) {
    xmlrpc_value *result;
    const char *resultString;

    LOG_Write(DEBUG,"Making XMLRPC call to server url '%s' method '%s'", serverUrl, methodName);

    /* Make the remote procedure call */
    result = xmlrpc_client_call(env, serverUrl, methodName,"(s)", methodArgs);
    if(env->fault_occurred) {
        LOG_Write(ERROR, "XMLRPC_Query: %s (%d)\n", env->fault_string, env->fault_code);
        return NULL;
    }

    xmlrpc_read_string(env, result, &resultString);
    if(env->fault_occurred) {
        LOG_Write(ERROR, "XMLRPC_Query: %s (%d)\n", env->fault_string, env->fault_code);
        return NULL;
    }

    /* Dispose of our result value. */
    xmlrpc_DECREF(result);
    return resultString;
}

int XMLRPC_Close(xmlrpc_env *env) {
    /* Clean up our error-handling environment. */
    xmlrpc_env_clean(env);
    /* Shutdown our XML-RPC client library. */
    xmlrpc_client_cleanup();
}

#else
#warning XMLRPC disabled
#endif
